/*
 * Copyright (c) 2005 The University of Utah
 *
 * This file is part of Jiazzi.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions, and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package out;

import num.Num;

/**
 * @author eeide
 *
 */
public class Evaluator implements ExprVisitor {
    // static protected final Expr BAD_ARG
    //     = new EvalErrorExpr("internal error (bad argument)");
    // static protected final Expr BAD_OP
    //     = new EvalErrorExpr("internal error (unknown operator)");
    protected static final Expr NO_MEM
        = new EvalErrorExpr("out of memory");
    protected static final Expr NO_CHILD
        = new EvalErrorExpr("internal error (null subexpression)");
    protected static final Expr MATH_ERROR
        = new EvalErrorExpr("math error");
    protected static final Expr DIV_BY_ZERO
        = new EvalErrorExpr("division by zero");

    public Evaluator() {
        super();
    }

    public Expr eval(Expr e) {
        return e.accept(this);
    }

    protected class PartialResult {
        public final Expr answer;
        public final Num operand1;
        public final Num operand2;

        public PartialResult(Expr e) {
            answer = e;
            operand1 = null;
            operand2 = null;
        }

        public PartialResult(Num n) {
            answer = null;
            operand1 = n;
            operand2 = null;
        }

        public PartialResult(Num n1, Num n2) {
            answer = null;
            operand1 = n1;
            operand2 = n2;
        }
    }

    protected PartialResult processUnary(UnaryExpr e) {
        Expr inner = e.getInner();
        if (inner == null) {
            return new PartialResult(NO_CHILD);
        }
        Expr innerResult = inner.accept(this);
        if (!(innerResult instanceof NumExpr)) {
            return new PartialResult(innerResult);
        }

        Num innerN = ((NumExpr) innerResult).getNum();
        return new PartialResult(innerN);
    }

    protected PartialResult processBinary(BinaryExpr e) {
        Expr left = e.getLeft();
        if (left == null) {
            return new PartialResult(NO_CHILD);
        }
        Expr leftResult = left.accept(this);
        if (!(leftResult instanceof NumExpr)) {
            return new PartialResult(leftResult);
        }

        Expr right = e.getRight();
        if (right == null) {
            return new PartialResult(NO_CHILD);
        }
        Expr rightResult = right.accept(this);
        if (!(rightResult instanceof NumExpr)) {
            return new PartialResult(rightResult);
        }

        Num leftN = ((NumExpr) leftResult).getNum();
        Num rightN = ((NumExpr) rightResult).getNum();
        return new PartialResult(leftN, rightN);
    }

    public Expr visitDifference(DifferenceExpr e) {
        PartialResult pe = processBinary(e);
        if (pe.answer != null) {
            return pe.answer;
        }
        Num n = pe.operand1.makeDifference(pe.operand2);
        if (n == null) {
            return MATH_ERROR;
        }
        return new NumExpr(n);
    }

    public Expr visitEvalError(EvalErrorExpr e) {
        return e;
    }

    public Expr visitNegative(NegativeExpr e) {
        PartialResult pe = processUnary(e);
        if (pe.answer != null) {
            return pe.answer;
        }
        Num n = pe.operand1.makeNegate();
        if (n == null) {
            return MATH_ERROR;
        }
        return new NumExpr(n);
    }

    public Expr visitNum(NumExpr e) {
        return e;
    }

    public Expr visitPositive(PositiveExpr e) {
        PartialResult pe = processUnary(e);
        if (pe.answer != null) {
            return pe.answer;
        }
        Num n = pe.operand1;
        if (n == null) {
            return MATH_ERROR;
        }
        return new NumExpr(n);
    }

    public Expr visitProduct(ProductExpr e) {
        PartialResult pe = processBinary(e);
        if (pe.answer != null) {
            return pe.answer;
        }
        Num n = pe.operand1.makeProduct(pe.operand2);
        if (n == null) {
            return MATH_ERROR;
        }
        return new NumExpr(n);
    }

    public Expr visitRatio(RatioExpr e) {
        PartialResult pe = processBinary(e);
        if (pe.answer != null) {
            return pe.answer;
        }
        Num n = pe.operand1.makeRatio(pe.operand2);
        if (n == null) {
            if (pe.operand2.isZero()) {
                return DIV_BY_ZERO;
            }
            return MATH_ERROR;
        }
        return new NumExpr(n);
    }

    public Expr visitSum(SumExpr e) {
        PartialResult pe = processBinary(e);
        if (pe.answer != null) {
            return pe.answer;
        }
        Num n = pe.operand1.makeSum(pe.operand2);
        if (n == null) {
            return MATH_ERROR;
        }
        return new NumExpr(n);
    }

    public Expr visitTokenError(TokenErrorExpr e) {
        return e;
    }
}
