/*
 * Copyright (c) 2005 The University of Utah
 *
 * This file is part of Jiazzi.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions, and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package out;

import java.io.PrintStream;

/**
 * @author eeide
 *
 */
public class Printer implements ExprVisitor {
    private static final int EXPR_PREC_INIT = -1;
    private static final int EXPR_PREC_SUM = 0;
    private static final int EXPR_PREC_PRODUCT = 1;
    private static final int EXPR_PREC_UNARY = 2;

    private final PrintStream out;

    protected int prec;

    public Printer(PrintStream ps) {
        super();
        out = ps;
    }

    public void print(Expr e) {
        // A low EXPR_PREC_INIT value lets us avoid any outermost ()'s.
        prec = EXPR_PREC_INIT;
        e.accept(this);
    }

    protected void processUnary(UnaryExpr expr, char exprOp, int exprPrec) {
        if (prec > exprPrec) {
            out.print('(');
        }
        out.print(exprOp);
        int savePrec = prec;
        prec = exprPrec;
        expr.getInner().accept(this);
        prec = savePrec;
        if (prec > exprPrec) {
            out.print(')');
        }
    }

    protected void processBinary(BinaryExpr expr, char exprOp, int exprPrec) {
        if (prec >= exprPrec) {
            out.print('(');
        }
        int savePrec = prec;
        prec = exprPrec - 1;
        expr.getLeft().accept(this);
        out.print(' ');
        out.print(exprOp);
        out.print(' ');
        prec = exprPrec;
        expr.getRight().accept(this);
        prec = savePrec;
        if (prec >= exprPrec) {
            out.print(')');
        }
    }

    /* (non-Javadoc)
     * @see expr.ExprVisitor#visitDifference(expr.DifferenceExpr)
     */
    public Expr visitDifference(DifferenceExpr e) {
        processBinary(e, '-', EXPR_PREC_SUM);
        return null;
    }

    /* (non-Javadoc)
     * @see expr.ExprVisitor#visitEvalError(expr.ErrorExpr)
     */
    public Expr visitEvalError(EvalErrorExpr e) {
        out.print("error: ");
        out.print(e.getString());
        return null;
    }

    /* (non-Javadoc)
     * @see expr.ExprVisitor#visitNegative(expr.NegativeExpr)
     */
    public Expr visitNegative(NegativeExpr e) {
        processUnary(e, '-', EXPR_PREC_UNARY);
        return null;
    }

    /* (non-Javadoc)
     * @see expr.ExprVisitor#visitNum(expr.NumExpr)
     */
    public Expr visitNum(NumExpr e) {
        out.print(e.getNum().toString());
        return null;
    }

    /* (non-Javadoc)
     * @see expr.ExprVisitor#visitPositive(expr.PositiveExpr)
     */
    public Expr visitPositive(PositiveExpr e) {
        processUnary(e, '+', EXPR_PREC_UNARY);
         return null;
    }

    /* (non-Javadoc)
     * @see expr.ExprVisitor#visitProduct(expr.ProductExpr)
     */
    public Expr visitProduct(ProductExpr e) {
        processBinary(e, '*', EXPR_PREC_PRODUCT);
        return null;
    }

    /* (non-Javadoc)
     * @see expr.ExprVisitor#visitRatio(expr.RatioExpr)
     */
    public Expr visitRatio(RatioExpr e) {
        processBinary(e, '/', EXPR_PREC_PRODUCT);
        return null;
    }

    /* (non-Javadoc)
     * @see expr.ExprVisitor#visitSum(expr.SumExpr)
     */
    public Expr visitSum(SumExpr e) {
        processBinary(e, '+', EXPR_PREC_SUM);
        return null;
    }

    /* (non-Javadoc)
     * @see expr.ExprVisitor#visitTokenError(expr.TokenErrorExpr)
     */
    public Expr visitTokenError(TokenErrorExpr e) {
        out.print("scanner error: ");
        out.print(e.getToken().toString());
        return null;
    }
}
