/*
 * Copyright (c) 2005 The University of Utah
 *
 * This file is part of Jiazzi.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions, and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package out;

import expr.DifferenceExpr;
import expr.ErrorExpr;
import expr.Expr;
import expr.NegativeExpr;
import expr.NumExpr;
import expr.PositiveExpr;
import expr.ProductExpr;
import expr.RatioExpr;
import expr.SumExpr;
import expr.TokenErrorExpr;
import token.EOFToken;
import token.LParenToken;
import token.MinusToken;
import token.NumToken;
import token.PlusToken;
import token.RParenToken;
import token.SlashToken;
import token.StarToken;
import token.Token;
import token.Tokenizer;

/**
 * @author eeide
 *
 */
public class Parser {
    /**
     *
     */
    protected Tokenizer tokenizer;

    /**
     *
     */
    public Parser() {
        super();
    }

    /**
     * @param s
     * @return
     */
    public Expr parse(String s) {
        tokenizer = new Tokenizer(s);

        Expr e = parseSum();
        if (e instanceof ErrorExpr) {
            return e;
        }
        Token t = tokenizer.nextToken();
        if (!(t instanceof EOFToken)) {
            return new TokenErrorExpr(t);
        }
        return e;
    }

    /**
     * @param op
     * @param inner
     * @return
     */
    protected Expr makeUnaryExpr(Token op, Expr inner) {
        if (op instanceof PlusToken) {
            return new PositiveExpr(inner);
        }
        if (op instanceof MinusToken) {
            return new NegativeExpr(inner);
        }
        return new TokenErrorExpr(op);
    }

    /**
     * @param op
     * @param lhs
     * @param rhs
     * @return
     */
    protected Expr makeBinaryExpr(Token op, Expr lhs, Expr rhs) {
        if (op instanceof PlusToken) {
            return new SumExpr(lhs, rhs);
        }
        if (op instanceof MinusToken) {
            return new DifferenceExpr(lhs, rhs);
        }
        if (op instanceof StarToken) {
            return new ProductExpr(lhs, rhs);
        }
        if (op instanceof SlashToken) {
            return new RatioExpr(lhs, rhs);
        }
        return new TokenErrorExpr(op);
    }

    /**
     * @return
     */
    protected Expr parseSum() {
        Expr lhs;

        lhs = null;
        for (;;) {
            // If needed, parse the left hand side of the (binary) sum.
            if (lhs == null) {
                lhs = parseProduct();
            }
            if (lhs instanceof ErrorExpr) {
                return lhs;
            }

            // Parse the operator.
            Token op = tokenizer.nextToken();
            if ((op instanceof PlusToken) || (op instanceof MinusToken)) {
                // Expected case: nothing to do.
            } else {
                // Push the token back onto the input...
                tokenizer.pushToken(op);
                // ...and return the parsed sum.
                return lhs;
            }

            // Parse the right hand side of the (binary) sum.
            Expr rhs = parseProduct();
            if (rhs instanceof ErrorExpr) {
                return rhs;
            }

            // Construct the parsed expression.  This becomes the left hand
            // side for the next iteration through the loop.
            lhs = makeBinaryExpr(op, lhs, rhs);
        }
    }

    /**
     * @return
     */
    protected Expr parseProduct() {
        Expr lhs;

        lhs = null;
        for (;;) {
            // If needed, parse the left hand side of the product.
            if (lhs == null) {
                lhs = parseUnary();
            }
            if (lhs instanceof ErrorExpr) {
                return lhs;
            }

            // Parse the operator.
            Token op = tokenizer.nextToken();
            if ((op instanceof StarToken) || (op instanceof SlashToken)) {
                // Expected case: nothing to do.
            } else {
                // Push the token back onto the input...
                tokenizer.pushToken(op);
                // ...and return the parsed product.
                return lhs;
            }

            // Parse the right hand side of the product.
            Expr rhs = parseUnary();
            if (rhs instanceof ErrorExpr) {
                return rhs;
            }

            // Construct the parsed expression.  This becomes the left hand
            // side for the next iteration through the loop.
            lhs = makeBinaryExpr(op, lhs, rhs);
        }
    }

    /**
     * @return
     */
    protected Expr parseUnary() {
        // Parse the operator.
        Token op = tokenizer.nextToken();
        if ((op instanceof PlusToken) || (op instanceof MinusToken)) {
            // Expected case: nothing to do.
        } else {
            // Push the token back onto the input...
            tokenizer.pushToken(op);
            // ...and parse and return the underlying term.
            return parseTerm();
        }

        // Recurse: parse the inner unary expression,
        Expr inner = parseUnary();
        if (inner instanceof ErrorExpr) {
            return inner;
        }

        // Return the unary expr applied to the inner expr.
        return makeUnaryExpr(op, inner);
    }

    /**
     * @return
     */
    protected Expr parseTerm() {
        Token t = tokenizer.nextToken();

        // Easy case: create and return a number expression.
        if (t instanceof NumToken) {
            return new NumExpr(((NumToken) t).getNum());
        }

        // Another easy case: parse a parenthesized sum.
        if (t instanceof LParenToken) {
            Expr sum = parseSum();
            if (sum instanceof ErrorExpr) {
                return sum;
            }
            t = tokenizer.nextToken();
            if (t instanceof RParenToken) {
                return sum;
            }
            return new TokenErrorExpr(t);
        }

        // A third easy case: anything else is an error.
        return new TokenErrorExpr(t);
    }
}
