/*
 * Copyright (c) 2005 The University of Utah
 *
 * This file is part of Jiazzi.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions, and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package out;

import java.util.NoSuchElementException;
import java.util.StringTokenizer;

import num.Num;

/**
 * @author eeide
 *
 */
public class Tokenizer {
    protected static final ErrorToken BAD_ARG
        = new ErrorToken("internal error (bad argument)");
    protected static final ErrorToken BAD_INPUT
        = new ErrorToken("unrecognized input");
    // protected static final ErrorToken NO_MEM
    //     = new ErrorToken("out of memory");

    private static final EOFToken SCAN_EOF
        = new EOFToken();

    private final StringTokenizer input;

    private Token savedToken;

    public Tokenizer(String s) {
        super();
        if (s == null) {
            savedToken = BAD_ARG;
            s = "";
        }
        input = new StringTokenizer(s, " \t\n\r\f+-*/()", true);
    }

    public Token nextToken() {
        if (savedToken != null) {
            Token t = savedToken;
            savedToken = null;
            return t;
        }

        String s;
        char c;

        // Skip forward over whitespace.  Signal EOF if we run out of input.
        do {
            try {
                s = input.nextToken();
            } catch (NoSuchElementException nse) {
                return SCAN_EOF;
            }
            // The token string should always contain at least one character,
            // but play it safe anyway.
            if (s.length() >= 1) {
                c = s.charAt(0);
            } else {
                c = ' ';
            }
        } while (Character.isWhitespace(c));

        // Recognize symbols.
        if (s.length() == 1) {
            switch (c) {
            case '+':
                return new PlusToken();
            case '-':
                return new MinusToken();
            case '*':
                return new StarToken();
            case '/':
                return new SlashToken();
            case '(':
                return new LParenToken();
            case ')':
                return new RParenToken();
            default:
                break;
            }
        }

        // Recognize numbers.  Delegate to the Num class.
        Num n = Num.maybeNum(s);
        if (n != null) {
            return new NumToken(n);
        }

        // Otherwise, it's a scanner error.
        return BAD_INPUT;
    }

    public void pushToken(Token t) {
        savedToken = t;
    }
}
